#include "fdb_writer.hpp"

#include <eccodes.h>
#include <eckit/runtime/Main.h>
#include <fdb5/api/FDB.h>
#include <fdb5/config/Config.h>
#include <yaco/chrono.hpp>
#include <yaco/collection.hpp>
#include <yaco/config.hpp>
#include <yaco/grid.hpp>

using yaco::FDBWriter;
using yaco::Grid;
using yaco::chrono::DateTime;

namespace config = yaco::config;

std::unique_ptr<yaco::DataHandler> yaco::DataHandler::createFDBWriter(const config::Value& doc)
{
    return std::make_unique<FDBWriter>(doc);
}

class FDBWriter::GRIBMessage {
  private:
    codes_handle* handle_ = nullptr;

  public:
    GRIBMessage()
    {
        handle_ = codes_grib_handle_new_from_samples(0, "GRIB2");
        if (handle_ == nullptr) {
            throw std::runtime_error("Failed to create eccodes handle");
        }
    }
    ~GRIBMessage()
    {
        codes_handle_delete(handle_);
    }

    void set(const DateTime& dt)
    {
        GRIB_CHECK(codes_set_long(handle_, "dataDate", dt.year() * 10000 + dt.month() * 100 + dt.day()), 0);
        GRIB_CHECK(codes_set_long(handle_, "hour", dt.hour()), 0);
        GRIB_CHECK(codes_set_long(handle_, "minute", dt.minute()), 0);
        GRIB_CHECK(codes_set_long(handle_, "second", dt.second()), 0);
    }

    void set(const Field& field)
    {
        CODES_CHECK(codes_set_double_array(handle_, "values", field.data(), field.size()), 0);
    }

    void set(const std::string& key, const std::string& value)
    {
        std::size_t len = value.size();
        int res         = codes_set_string(handle_, key.c_str(), value.c_str(), &len);
        if (res != 0) {
            throw std::runtime_error("Failed to set key '" + key + "' to '" + value + "': " + codes_get_error_message(res));
        }
    }

    void set(const std::string& key, double value)
    {
        int res = codes_set_double(handle_, key.c_str(), value);
        if (res != 0) {
            throw std::runtime_error("Failed to set key '" + key + "' to '" + std::to_string(value) + "': " + codes_get_error_message(res));
        }
    }

    void archive(fdb5::FDB& fdb, const fdb5::Key& key)
    {
        const void* data;
        std::size_t data_length;
        int res = codes_get_message(handle_, &data, &data_length);
        if (res != 0) {
            throw std::runtime_error("Failed to get raw message: " + std::string(codes_get_error_message(res)));
        }
        fdb.archive(key, data, data_length);
    }
};

class FDBWriter::Impl {
  public:
    Impl(const config::Value& config) : config_(config)
    {
        if (!fdb_lib_initialized_) {
            std::string arg = "fdb";
            std::array<char*, 2> argv({ arg.data(), nullptr });
            eckit::Main::initialise(argv.size() - 1, argv.data());
            fdb_lib_initialized_ = true;
        }

        const auto fdb_config = config_.find("fdb_config");
        if (fdb_config != config_.mapEnd()) {
            fdb_ = std::make_unique<fdb5::FDB>(fdb5::Config::make(fdb_config->second.asString()));
        } else {
            fdb_ = std::make_unique<fdb5::FDB>();
        }
    }

    void handle(Collection& collection, const DateTime& timestep);

  private:
    static bool fdb_lib_initialized_;
    config::Value config_;
    std::unique_ptr<fdb5::FDB> fdb_;
};

bool FDBWriter::Impl::fdb_lib_initialized_ = false;

FDBWriter::FDBWriter(const config::Value& config)
        : impl_(std::make_unique<Impl>(config))
{
}

FDBWriter::~FDBWriter() = default;

template<typename T>
void extract_string_map(const config::Value& value, const std::string& key, T& dict, const DateTime& datetime, const Grid& grid, size_t collection_index)
{
    const auto keyIter = value.find(key);
    if (keyIter != value.mapEnd()) {
        for (auto iter = keyIter->second.mapBegin(); iter != keyIter->second.mapEnd(); ++iter) {
            if (iter->second.isMap()) {
                const auto use = iter->second.at("use").asString();
                if (use == "datetime") {
                    dict.set(iter->first, datetime.toString(iter->second.at("datetime").asString().c_str()));
                } else if (use == "grid") {
                    dict.set(iter->first, grid.name());
                } else if (use == "collection_index") {
                    dict.set(iter->first, std::to_string(static_cast<long>(collection_index) + (iter->second.has("add") ? iter->second.at("add").asInt32() : 0)));
                } else {
                    throw std::runtime_error("Unknown use type '" + use + "'");
                }
            } else {
                dict.set(iter->first, iter->second.asString());
            }
        }
    }
}

void FDBWriter::Impl::handle(Collection& collection, const DateTime& datetime)
{
    const auto& collection_node = config_.at("collections").at(collection.name());

    config::Value::ConstArrayIterator node_it;
    bool per_collection_member = collection_node.has("collection");
    if (per_collection_member) {
        node_it = collection_node.at("collection").arrayBegin();
        if (collection_node.at("collection").asArray().size() != collection.size()) {
            throw std::runtime_error("Collection size mismatch");
        }
    }

    for (std::size_t i = 0; i < collection.size(); ++i) {
        const auto& field = collection.at(i);

        GRIBMessage msg;
        msg.set(datetime);
        msg.set(field);
        extract_string_map(config_, "grib", msg, datetime, collection.grid(), i);
        extract_string_map(collection_node, "grib", msg, datetime, collection.grid(), i);
        if (per_collection_member) {
            extract_string_map(*node_it, "grib", msg, datetime, collection.grid(), i);
        }

        fdb5::Key key{};
        extract_string_map(config_, "fdb_key", key, datetime, collection.grid(), i);
        extract_string_map(collection_node, "fdb_key", key, datetime, collection.grid(), i);
        if (per_collection_member) {
            extract_string_map(*node_it, "fdb_key", key, datetime, collection.grid(), i);
        }
        msg.archive(*fdb_, key);

        if (per_collection_member) {
            ++node_it;
        }
    }
}

void FDBWriter::handle(Collection& collection, const DateTime& datetime)
{
    impl_->handle(collection, datetime);
}
