#include "healpix_grid.hpp"

#include <healpix_cxx/healpix_base.h>
#include <mpi.h>
#include <numeric>
#include <optional>

extern "C" {
#include <yac_config.h>
#include <yac_interface.h>
}

using yaco::Grid;
using yaco::HealpixGrid;

static std::optional<int> searchIndex(const std::vector<double>& lon, const std::vector<double>& lat, double lon_value, double lat_value);

static void getHealpixGridData(int sides, Grid::HealpixScheme scheme, int& nbrCells, int& nbrVertices, std::vector<int>& nbrVerticesPerCell,
                               std::vector<int>& cellToVertex,
                               std::vector<double>& verticesLon,
                               std::vector<double>& verticesLat,
                               std::vector<double>& centersLon,
                               std::vector<double>& centersLat);

std::unique_ptr<Grid> Grid::createHealpix(std::string name, std::size_t sides, HealpixScheme scheme)
{
    return std::make_unique<HealpixGrid>(name, sides, scheme);
}

HealpixGrid::HealpixGrid(std::string name, std::size_t sides, HealpixScheme scheme)
        : name_(name)
{
    getHealpixGridData(static_cast<int>(sides), scheme, nbrCells_, nbrVertices_, nbrVerticesPerCell_,
                       cellToVertex_, verticesLon_, verticesLat_,
                       centersLon_, centersLat_);
}

static std::optional<int> searchIndex(const std::vector<double>& lon, const std::vector<double>& lat, double lon_value, double lat_value)
{
    auto iter = std::find(lon.begin(), lon.end(), lon_value);
    if (iter == lon.end()) {
        return std::nullopt;
    }
    std::size_t index = static_cast<std::size_t>(iter - lon.begin());
    if (index >= lat.size()) {
        return std::nullopt;
    }
    if (lat[index] == lat_value) {
        return static_cast<int>(index);
    }
    return std::nullopt;
}

static void getHealpixGridData(int sides, Grid::HealpixScheme scheme, int& nbrCells, int& nbrVertices,
                               std::vector<int>& nbrVerticesPerCell,
                               std::vector<int>& cellToVertex,
                               std::vector<double>& verticesLon,
                               std::vector<double>& verticesLat,
                               std::vector<double>& centersLon,
                               std::vector<double>& centersLat)
{
    enum Healpix_Ordering_Scheme healpixScheme;
    switch (scheme) {
        case Grid::HealpixScheme::ring:
            healpixScheme = RING;
            break;
        case Grid::HealpixScheme::nested:
            healpixScheme = NEST;
            break;
        default:
            throw std::runtime_error("Unknown Healpix scheme");
            break;
    }

    Healpix_Base hb(sides, healpixScheme);
    nbrCells = hb.Npix();
    nbrVerticesPerCell.resize(static_cast<std::size_t>(nbrCells), 4);
    for (int i = 0; i < nbrCells; ++i) {
        {
            pointing p = hb.pix2ang(i);
            centersLon.push_back(p.phi);
            centersLat.push_back(p.theta);
        }

        std::vector<vec3> corners;
        hb.boundaries(i, 1, corners);

        for (const vec3& v : corners) {
            pointing p;
            p.from_vec3(v);
            double lon        = p.phi;
            double lat        = (M_PI / 2) - p.theta;
            auto searchResult = searchIndex(verticesLon, verticesLat, lon, lat);
            if (searchResult.has_value()) {
                cellToVertex.push_back(searchResult.value());
            } else {
                verticesLon.push_back(lon);
                verticesLat.push_back(lat);
                cellToVertex.push_back(static_cast<int>(verticesLon.size() - 1));
            }
        }
        nbrVertices = static_cast<int>(verticesLon.size());
    }
}
