#include "yac2_input.hpp"

#include "yac2_grid_definer.hpp"

#include <cmath>
#include <mpi.h>
#include <set>
#include <stdexcept>
#include <vector>
#include <yaco/chrono.hpp>
#include <yaco/collection.hpp>
#include <yaco/config.hpp>

extern "C" {
#include <yac_config.h>
#include <yac_interface.h>
}

using yaco::Yac2Input;
using yaco::YacComponent;
using yaco::chrono::DateTime;
using yaco::chrono::Duration;

#ifndef YACO_DEFAULT_MISSING_VALUE
#define YACO_DEFAULT_MISSING_VALUE -9.e33
#endif

std::unique_ptr<YacComponent>
YacComponent::createYac2Input(const config::Value& doc, std::unique_ptr<Grid> grid, std::unique_ptr<DataHandler> handler)
{
    return std::make_unique<Yac2Input>(doc, std::move(grid), std::move(handler));
}

Yac2Input::Yac2Input(const config::Value& config, std::unique_ptr<Grid> grid, std::unique_ptr<DataHandler> handler)
        : name_(config.at("name").asString()), grid_(std::move(grid)), handler_(std::move(handler))
{
    yac_cinit(config.at("yac_xml").asString().c_str(), config.at("yac_xsd").asString().c_str());
    yac_cdef_comp(name_.c_str(), &id_);

    for (const auto& collection : config.at("collections").asArray()) {
        collections_.emplace_back(collection.at("name").asString(), static_cast<std::size_t>(collection.at("size").asInt32()), *grid_);
    }
    timeStep_  = config.at("time_step").asString();
    timeLag_   = config.has("time_lag") ? config.at("time_lag").asInt32() : 0;
    startDate_ = config.at("start_date").asString();
    endDate_   = config.at("end_date").asString();
}

Yac2Input::~Yac2Input()
{
    yac_cfinalize();
}

void Yac2Input::run()
{
    Yac2GridDefiner definer;
    int gridId = grid_->accept(definer);

    std::vector<int> collectionId(collections_.size(), 0);
    for (std::size_t i = 0; i < collections_.size(); ++i) {
        yac_cdef_field(collections_[i].name().c_str(),
                       id_,
                       &gridId,
                       1,
                       &collectionId[i]);
    }

    // start coupling
    {
        int ierror = 0;
        yac_csearch(&ierror);
        if (ierror != 0) {
            throw std::runtime_error("Error " + std::to_string(ierror) + " in yac_csearch");
        }
    }

    DateTime startTime(startDate_);
    DateTime endTime(endDate_);

    Duration baseTimeStep(timeStep_);
    DateTime& step = startTime;
    for (int i = 0; i < timeLag_; ++i) {
        step += baseTimeStep;
    }
    for (; step < endTime; step += baseTimeStep) {
        for (std::size_t collectionIdx = 0; collectionIdx < collections_.size(); ++collectionIdx) {
            std::vector<double*> buffers(collections_[collectionIdx].size(), nullptr);
            for (std::size_t fieldIdx = 0; fieldIdx < collections_[collectionIdx].size(); ++fieldIdx) {
                buffers[fieldIdx] = collections_[collectionIdx].at(fieldIdx).data();
            }

            int info   = 0;
            int ierror = 0;
            yac_cget(collectionId[collectionIdx], static_cast<int>(collections_[collectionIdx].size()), buffers.data(), &info, &ierror);

            if (ierror != 0) {
                throw std::runtime_error("Error " + std::to_string(ierror) + " in yac_cget");
            }

            if (info != YAC_ACTION_NONE) {
                // Reset NANs to missing value
                for (std::size_t fieldIdx = 0; fieldIdx < collections_[collectionIdx].size(); ++fieldIdx) {
                    for (std::size_t i = 0; i < collections_[collectionIdx].at(fieldIdx).size(); ++i) {
                        if (std::isnan(collections_[collectionIdx].at(fieldIdx).at(i))) {
                            collections_[collectionIdx].at(fieldIdx).at(i) = YACO_DEFAULT_MISSING_VALUE;
                        }
                    }
                }

                handler_->handle(collections_[collectionIdx], step);
            }
        }
    }
}
